;;; Foreign function interfaces for FFTW version 3.

;;; Copyright (C) 2009 Joern Inge Vestgaarden
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License along
;;; with this program; if not, write to the Free Software Foundation, Inc.,
;;; 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

;;; TODO: the calls should be wrapped in unwind protect 
;;;       to avoid memory leaks 

(in-package :fftw-ffi)

(defconstant +double-float-bytes+ (truncate (sb-alien:ALIEN-SIZE sb-alien:double-float) 8))

(defconstant +FFTW-ESTIMATE+ 64)

(defconstant +FFTW-FORWARD+ -1)

(defconstant +FFTW-BACKWARD+ 1)

(declaim (inline |fftw_plan_dft_1d|))
(define-alien-routine |fftw_plan_dft_1d|
    (* t)
  (n int)
  (in (* double-float))
  (out (* double-float))
  (sign int)
  (flags int))
 
(declaim (inline |fftw_plan_dft_2d|))
(define-alien-routine |fftw_plan_dft_2d|
    (* t)
  (n0 int)
  (n1 int)
  (in (* double-float))
  (out (* double-float))
  (sign int)
  (flags int))

(declaim (inline |fftw_execute|))
(define-alien-routine |fftw_execute| 
    void
  (plan (* t)))

(declaim (inline |fftw_destroy_plan|))
(define-alien-routine |fftw_destroy_plan| 
    void 
  (plan (* t)))

(defun fftw-fft1 (n a astart b bstart direction flag)
  "One dimensional fft by forign call to fftw."
  ;; TODO we should handle conditions to avoid mem-leaks
  (let ((astart (* astart +double-float-bytes+))
	(bstart (* bstart +double-float-bytes+)))
    (with-pinned-objects (a b)  
      (let ((plan (|fftw_plan_dft_1d| 
		   n  
		   (sap+ (vector-sap a) astart) 
		   (sap+ (vector-sap b) bstart)
		   direction
		   flag)))
	(|fftw_execute| plan)
	(|fftw_destroy_plan| plan)))
    b))

(defun fftw-fft2 (m n in out direction flag)
  "Two dimensional fft by forign call to fftw."
  ;; TODO we should handle conditions to avoid mem-leaks
  (with-pinned-objects (in out)  
    (let ((plan (|fftw_plan_dft_2d| 
		 n ; swap n and m due to row major order
		 m
		 (vector-sap in)
		 (vector-sap out)		 
		 direction
		 flag)))
      (|fftw_execute| plan)
      (|fftw_destroy_plan| plan)))
  out)


;;;; Now multi-thread code

(declaim (inline |fftw_init_threads|))
(define-alien-routine |fftw_init_threads|
    int)
  
(declaim (inline |fftw_plan_with_nthreads|))
(define-alien-routine |fftw_plan_with_nthreads|
    void
  (nthreads int))

(declaim (inline |fftw_cleanup_threads|))
(define-alien-routine |fftw_cleanup_threads|
    void)

(defun fftw-init-threads (num-threads)
  ;; Note: assumes that |fftw_init_threads| has been called!
  (|fftw_plan_with_nthreads| num-threads))

(defun fftw-cleanup-threads ()
  (|fftw_cleanup_threads|))

