;;; Level3-fft-fftw.lisp, fast fourier by fftw. 

;;; Copyright (C) 2009 Joern Inge Vestgaarden
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License along
;;; with this program; if not, write to the Free Software Foundation, Inc.,
;;; 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

;;; There is a sligh missuse of notation here, since the method 
;;; specialize on matrix-blas-zge, rather than matrix-fftw-zge, which
;;; would have a more correct name.

(in-package :lisplab)

(defmethod fft1 ((x matrix-blas-dge) &key)
  (fft1 (convert-to-matrix-zge x)))

(defmethod ifft1 ((x matrix-blas-dge) &key)
  (ifft1 (convert-to-matrix-zge x)))

(defmethod fft2 ((x matrix-blas-dge) &key)
  (fft2 (convert-to-matrix-zge x)))

(defmethod ifft2 ((x matrix-blas-dge) &key)
  (ifft2 (convert-to-matrix-zge x)))

(defun use-fftw-p ()
  cl-user::*lisplab-libfftw-path*)

(defun fftw-use-threads-p ()
  (and cl-user::*lisplab-libfftw-threads-path*
       (> *lisplab-num-threads* 1)))

(defmethod init-threads :after (num-threads)
  (when (fftw-use-threads-p)
    (fftw-ffi:fftw-init-threads num-threads)))

(defmethod cleanup-threads :after ()
  (when (fftw-use-threads-p)
    (fftw-ffi:fftw-cleanup-threads)))

(defmethod fft1 ((x matrix-blas-zge) &key)
   (if (not (use-fftw-p))
      (call-next-method)
      (let* ((rows (rows x))
	     (cols (cols x))
	     (store-x (matrix-store x))
	     (y (mcreate x))
	     (store-y (matrix-store y)))	 
	(dotimes (i cols)
	  ;; Could be made in parallel
	  (fftw-ffi:fftw-fft1 
	   rows
	   store-x
	   (* i cols)
	   store-y
	   (* i cols)
	   fftw-ffi:+FFTW-FORWARD+
	   fftw-ffi:+FFTW-ESTIMATE+))
	y)))

(defmethod ifft1 ((x matrix-blas-zge) &key)
   (if (not (use-fftw-p))
      (call-next-method)
      (let* ((rows (rows x))
	     (cols (cols x))
	     (store-x (matrix-store x))
	     (y (mcreate x))
	     (store-y (matrix-store y)))	 
	(dotimes (i cols)
	  ;; Could be made in parallel
	  (fftw-ffi:fftw-fft1 
	   rows
	   store-x
	   (* i cols)
	   store-y
	   (* i cols)
	   fftw-ffi:+FFTW-BACKWARD+
	   fftw-ffi:+FFTW-ESTIMATE+))
	y)))

(defmethod fft2 ((x matrix-blas-zge) &key)
  (if (not (use-fftw-p))
      (call-next-method)
      (let ((y (mcreate x)))
	(fftw-ffi:fftw-fft2 
	 (rows x) 
	 (cols x) 
	 (matrix-store x) 
	 (matrix-store y) 
	 fftw-ffi:+fftw-forward+ 
	 fftw-ffi:+FFTW-ESTIMATE+)
	y)))

(defmethod ifft2 ((x matrix-blas-zge) &key)
  (if (not (use-fftw-p))
      (call-next-method)
      (let ((y (mcreate x)))
	(fftw-ffi:fftw-fft2 
	 (rows x) 
	 (cols x) 
	 (matrix-store x) 
	 (matrix-store y) 
	 fftw-ffi:+fftw-backward+ 
	 fftw-ffi:+FFTW-ESTIMATE+)
	y)))

;;; TODO: remove the destructive mothods below. They only mess things up

(defun fft1!-forward-or-backward (x direction)
  (let* ((rows (rows x))
	 (cols (cols x))
	 (store (matrix-store x)))
    (dotimes (i cols)
      (fftw-ffi:fftw-fft1   
       rows
       store
       (* i cols)
       store
       (* i cols)
       direction
       fftw-ffi:+FFTW-ESTIMATE+)))
  x)

(defmethod fft1! ((x matrix-blas-zge) &key)
  (if cl-user::*lisplab-libfftw-path*
      (fft1!-forward-or-backward x fftw-ffi:+fftw-forward+)
      (call-next-method)))

(defmethod ifft1! ((x matrix-blas-zge) &key)
  (if cl-user::*lisplab-libfftw-path*
      (fft1!-forward-or-backward x fftw-ffi:+fftw-backward+)
      (call-next-method)))

(defmethod fft2! ((x matrix-blas-zge) &key)
  (if cl-user::*lisplab-libfftw-path*
      (progn 
	(fftw-ffi:fftw-fft2 
	 (rows x) 
	 (cols x) 
	 (matrix-store x) 
	 (matrix-store x) 
	 fftw-ffi:+fftw-forward+ 
	 fftw-ffi:+FFTW-ESTIMATE+)
	x)
      (call-next-method)))

(defmethod ifft2! ((x matrix-blas-zge) &key)
  (if cl-user::*lisplab-libfftw-path*
      (progn
	(fftw-ffi:fftw-fft2 
	 (rows x) 
	 (cols x) 
	 (matrix-store x) 
	 (matrix-store x) 
	 fftw-ffi:+fftw-backward+ 
	 fftw-ffi:+FFTW-ESTIMATE+)
	x)
      (call-next-method)))
