;;; Lisplab, level1-util.lisp
;;; Level1, utility functions for matrix definitions.
;;; 

;;; Copyright (C) 2009 Joern Inge Vestgaarden
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License along
;;; with this program; if not, write to the Free Software Foundation, Inc.,
;;; 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

;;; TODO: change name of this to something about blas store
;;;
;;; This file contains manipulations of simple double-float arrays 
;;; and should be called by the spesialized matrix methods. 
;;; The purpose of this layer is that it can be used by 
;;; many classes such as matrix-base-dge and matrix-base-ddi, etc. 
;;; 
;;; The content of this file must be highly optimized 
;;; and should not depend anything exept Common Lisp itself.

(in-package :lisplab)

;;; Things that are common both for real and complex stores

(deftype type-blas-store ()
  '(simple-array double-float (*)))

#+(and :sbcl :x86) (deftype type-blas-idx ()
		     '(MOD #x1FFFFFFF))
#+(and :sbcl :x86-64) (deftype type-blas-idx ()
			'(MOD #xFFFFFFFFFFFFFFD))
#-:sbcl (deftype type-blas-idx ()
			'fixnum)

(declaim (inline column-major-idx))

(declaim (ftype (function 
		 (type-blas-idx
		  type-blas-idx
		  type-blas-idx)
		  type-blas-idx)
		column-major-idx))

(defun column-major-idx (i j rows)
  (truly-the type-blas-idx (+ i (truly-the type-blas-idx (* j rows)))))

(defun copy-matrix-stores (a b)
  (let ((len (length a)))
    (declare (type type-blas-store a b)
	     (type type-blas-idx len))
    (dotimes (i len)
      (setf (aref b i) (aref a i))))
  b)

;;;; The real store

(declaim (inline ref-blas-real-store (setf ref-blas-real-store)))

(declaim (ftype (function 
		 (type-blas-store
		  type-blas-idx
		  type-blas-idx
		  type-blas-idx)
		 double-float)
		ref-blas-real-store))

(defun ref-blas-real-store (store row col rows)
  "Matrix accessor for the real blas store"
  (aref (truly-the type-blas-store store)
	(truly-the type-blas-idx 
		   (column-major-idx (truly-the type-blas-idx row) 
				     (truly-the type-blas-idx col)
				     rows))))

(declaim (ftype (function 
		 (double-float
		   type-blas-store 
		   type-blas-idx
		   type-blas-idx
		   type-blas-idx
		  )
		 double-float)
		(setf ref-blas-real-store)))

(defun (setf ref-blas-real-store) (value store row col rows)
  (setf (aref (truly-the type-blas-store store)
	      (truly-the type-blas-idx 
			 (column-major-idx (truly-the type-blas-idx row) 
					   (truly-the type-blas-idx col)
					   rows)))
	value)
  value)

(defun allocate-real-store (size &optional (initial-element 0d0))
  ;; All matrix double and complex double constructors 
  ;; should call this one
  (let ((x (coerce initial-element 'double-float)))
    (declare (type double-float x)
	     (type type-blas-idx size))
    ;; Stupid efficiency hack for SBCL. Allocations of arrays with zeros 
    ;; is significantly faster than others!
    (if (= x 0d0) 	
	(make-array size
		    :element-type 'double-float
		    :initial-element 0d0)	
	(make-array size
		    :element-type 'double-float
		    :initial-element x))))

;;; Hopfully helpful iterators

(defmacro with-indices-df-1 (a m idx &body body)
  "Iterats over all the indices of one array"
  `(let ((,m ,a))
     (declare (type type-blas-store ,m))
     (dotimes (,idx (length ,m))
       (declare (type type-blas-idx ,idx)) 
       ,@body)))

(defmacro with-elements-df-1 (a elm &body body)
  "Iterats over all the elements of one array"
  (let ((m (gensym))
	(idx (gensym)))
    `(let ((,m ,a))
       (declare (type type-blas-store ,m))
       (dotimes (,idx (length ,m))
	 (declare (type type-blas-idx ,idx))
	 (let ((,elm (aref ,m ,idx)))
	   (declare (type double-float ,elm))       
	   ,@body)))))

;;;; The complex store

(defun allocate-complex-store (size &optional (value 0d0))
  (let* ((2size (* 2 size))
	 (rv (coerce (realpart value) 'double-float))
	 (iv (coerce (imagpart value) 'double-float))
	 (store (allocate-real-store 2size iv)))
    (declare (type type-blas-idx 2size)
	     (type double-float rv iv))
    (when (/= rv iv)    
      (loop for i from 0 below 2size by 2 do
	   (setf (aref store i) rv)))
    store))

(declaim (inline ref-blas-complex-store (setf ref-blas-complex-store)))

(declaim (ftype (function 
		 (type-blas-store
		  type-blas-idx
		  type-blas-idx
		  type-blas-idx)
		 (complex double-float))
		ref-blas-complex-store))

(defun ref-blas-complex-store (store row col rows)
  "Matrix accessor for the complet blas store"
  (let ((idx (truly-the type-blas-idx 
			(* 2 (column-major-idx (truly-the type-blas-idx row) 
					       (truly-the type-blas-idx col)
					       rows)))))    
    (declare (type-blas-idx idx))
    (complex (aref store idx)
	     (aref store (1+ idx)))))

(declaim (ftype (function 
		 ((complex double-float) 
		   type-blas-store 
		   type-blas-idx
		   type-blas-idx
		   type-blas-idx
		  )
		 (complex double-float))
		(setf ref-blas-complex-store)))

(defun (setf ref-blas-complex-store) (value store row col rows)
  (let ((idx (truly-the type-blas-idx 
			(* 2 (column-major-idx (truly-the type-blas-idx row) 
					       (truly-the type-blas-idx col)
					       rows)))))    
    (declare (type-blas-idx idx))
    (setf (aref store idx) (realpart value)
	  (aref store (1+ idx)) (imagpart value))
    value))

(declaim (ftype (function 
		 (type-blas-store
		  type-blas-idx)
		 (complex double-float))
		vref-blas-complex-store))

(defun vref-blas-complex-store (store idx)
  "Matrix accessor for the complex blas store"
  (let ((idx2 (truly-the type-blas-idx (* 2 idx))))
    (declare (type-blas-idx idx2)) 
    (complex (aref store idx2)
	     (aref store (1+ idx2)))))

(declaim (ftype (function 
		 ((complex double-float) 
		   type-blas-store 
		   type-blas-idx
		  )
		 (complex double-float))
		(setf vref-blas-complex-store)))

(defun (setf vref-blas-complex-store) (value store idx)
  (let ((idx2 (truly-the type-blas-idx (* 2 idx))))
    (declare (type-blas-idx idx2)) 
    (setf (aref store idx2) (realpart value)
	  (aref store (1+ idx2)) (imagpart value))
    value))

