;;; Lisplab, level2-constructors.lisp
;;; Possible and impossible ways to create matrices.

;;; Copyright (C) 2009 Joern Inge Vestgaarden
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License along
;;; with this program; if not, write to the Free Software Foundation, Inc.,
;;; 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

(in-package :lisplab)

(defmethod copy ((a matrix-base))
  (let ((x (make-matrix-instance (class-of a) (dim a) 0)))
    (dotimes (i (size x))
      (setf (vref x i) (vref a i)))
    x))

(defmethod mcreate ((a matrix-base) &optional (value 0) dim)
  (unless dim
    (setf dim (dim a)))
  (make-matrix-instance (class-of a) dim value)) 

(defmethod convert ((x matrix-base) type)
  (let ((y (make-matrix-instance type (dim x) 0)))
    (copy-contents x y)
    y))

(defun fill-matrix-with-list (m x)  
  "Helper function for convert."
  (let* ((rows (rows m))
	 (cols (cols m)))
    (do ((xx x (cdr xx))
	 (i 0 (1+ i)))
	((= i rows))
      (do ((yy (car xx) (cdr yy))
	   (j 0 (1+ j)))
	  ((= j cols))
	(setf (mref m i j) (car yy))))
    m))

(defmethod convert ((x cons) type)
  ;; Should it be moved to some other file?
  ;; TODO some better way ... some more general guessing routine 
  ;; like guess-best-element-type
  (if (consp (car x))
      (let* ((cols (length (car x)))
	     (rows (length x))
	     (m (make-matrix-instance type (list rows cols) 0)))
	(fill-matrix-with-list m x)
	m)
      ;; else make a row vector
      (convert (list x) type))) 

(defmethod mnew (type value rows &optional cols) 
  (make-matrix-instance type 
			(if cols 
			    (list rows cols)
			    rows) 
			value))

(defmacro mmat (type &body args)
  "Creates a matrix."
  `(convert 
    ,(cons 'list (mapcar (lambda (x) 
			   (cons 'list x)) 
			 args))
    ,type))

(defun mcol (type &rest args)
  "Creates a column matrix."
  (convert (mapcar 'list args) type))

(defun mrow (type &rest args)
  "Creates a row matrix."
  (convert args type))

;;; Constructors for matrix-dge

(defun drandom (rows cols)
  "Creates a double matrix with random element between 0 and 1."
  (mmap t #'random (dnew 1d0 rows cols)))

(defmacro dmat (&body args)
  "Creates a matrix-dge matrix."
  `(mmat 'matrix-dge ,@args))

(defun dcol (&rest args)
  "Creates a matrix-dge column matrix."
  (apply #'mcol 'matrix-dge args))

(defun drow (&rest args)
  "Creates a matrix-dge row matrix."
  (apply #'mrow 'matrix-dge args))

(defun dnew (value rows &optional (cols 1))
  "Creates a matrix-dge matrix"
  (mnew 'matrix-dge value rows cols))

(defun drange (n from to &optional (shift 0))
  "Creates a column vector of length n, with elements of equal spacing 
between from and to. The shift is the a number between 0 and 1 and 
shifts the start position.

For example: (drange 4 0 1) -> 0 1 2 3, while 
             (drange 4 0 1 0.5) -> 0.5 1.5 2.5 3.5."
  (let* ((x (dnew 0 n 1))
	 (store (matrix-store x))
	 (dx (/ (- (to-df to) (to-df from)) (to-df N)))
	 (shift (* dx (to-df shift))))
    (declare (type type-blas-store store)
	     (type type-blas-idx n)
	     (type double-float dx shift))
    (do ((i 0 (1+ i))
	 (v (to-df from) (+ v dx)))
	((>= i n) x)
      (declare (type double-float v))
      (setf (aref store i) (+ shift v)))))

(defun dgrid (xv yv)
  "Creates grid matrices from input vectors. Input are the x and y vectors 
and outputs are a list of x and y matrices. The input vectors are  
typically created with drange."
   (let* ((r (size xv))
	  (c (size yv))
	  (x (dnew 0 r c))
	  (y (dnew 0 r c))	  
	  (xv* (matrix-store xv))
	  (yv* (matrix-store yv))
	  (x* (matrix-store x))
	  (y* (matrix-store y)))
    (declare (type type-blas-store xv* yv* x* y*)
             (type type-blas-idx r c))
    (dotimes (i r)
      (dotimes (j c)    
        (let ((k (column-major-idx i j r)))
          (declare (type type-blas-idx k))
          (setf (aref x* k) (aref xv* i)
                (aref y* k) (aref yv* j)))))
    (list x y)))

;;; Constructors for matrix-zge

(defmacro zmat (&body args)
  "Creates a matrix-dge matrix."
  `(mmat 'matrix-zge ,@args))

(defun zcol (&rest args)
  "Creates a matrix-zge column matrix."
  (apply #'mcol 'matrix-zge args))

(defun zrow (&rest args)
  "Creates a matrix-zge row matrix."
  (apply #'mrow 'matrix-zge args))

(defun znew (value rows &optional (cols 1))
  "Creates a matrix-zge matrix"
  (mnew 'matrix-zge value rows cols))

(defun zgrid (xv yv)
  (let* ((r (size xv))
         (c (size yv))
         (x (znew 0 r c))
         (y (znew 0 r c)))
    (dotimes (i r)
      (dotimes (j c)
        (setf (mref x i j) (vref xv i)
              (mref y i j) (vref yv j))))
    (list x y)))


;;; Function matrix

(defun funmat (dim fun)
  "Creates a read only function matrix with column major order."
  (let ((rows (first dim))
	(cols (second dim)))
    (make-instance 'function-matrix 
		   :rows rows
		   :cols cols
		   :mref (lambda (self i j) 			   			   
			   (funcall fun i j))
		   :vref (lambda (self i)
			   (multiple-value-bind (c r)			       
			       (floor i rows)
			     (funcall fun r c))))))

(defun fmat (type dim fun)
  "Creates a matrix of of type type, dim dim from the function definition. 
Row major order"
  (convert (funmat dim fun)
	    type))


