;;; Lisplab, level2-matrix-dge.lisp
;;; Optimizations for real matrices.

;;; Copyright (C) 2009 Joern Inge Vestgaarden
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License along
;;; with this program; if not, write to the Free Software Foundation, Inc.,
;;; 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

(in-package :lisplab)

(defmethod mfill ((a matrix-base-dge) value)
  (let ((x (coerce value 'double-float))
	(store (matrix-store a)))
    (declare (type type-blas-store store))
    (fill store x)))

(defmethod copy ((matrix matrix-base-dge))
  (let ((store (matrix-store matrix)))
    (declare (type type-blas-store store))
    (make-instance (class-name (class-of matrix))
		   :store (copy-seq store)
		   :dim (dim matrix))))

(defmethod mreverse ((matrix matrix-base-dge))
  (let ((store (matrix-store matrix)))
    (declare (type type-blas-store store))
    (make-instance (class-name (class-of matrix))
		   :store (reverse store)
		   :dim (dim matrix))))

(defmethod copy-contents ((a matrix-base-dge) (b matrix-base-dge) &optional (converter nil))
  (let ((store-a (matrix-store a))
	(store-b (matrix-store b)))
    (declare (type type-blas-store store-a store-b))
    (if converter 
	(map-into store-b converter store-a)
	(copy-matrix-stores store-a store-b)))    
    b)

(defmethod mmap-into ((out matrix-base-dge) f (a matrix-base-dge) &rest args)
  (apply #'map-into 
	 (matrix-store out)
	 (lambda (&rest args)
	   (coerce (apply f args) 'double-float))
	 (matrix-store a) (mapcar #'matrix-store args))
    out)

(defmethod mmap ((type (eql nil)) f (a matrix-base-dge) &rest args)
  "No output. Called for side effects."
  (if args
      (apply #'map
	     nil
	     f 
	     (matrix-store a) (mapcar #'matrix-store args))
      (let ((store (matrix-store a)))
	(declare (type type-blas-store store))
	(map nil f store)))
  nil)

(defmethod msum ((m matrix-base-dge))
  (let ((sum 0d0))
    (with-elements-df-1 (matrix-store m) x
      (incf sum x))
    sum))

(defmethod mmax ((m matrix-base-dge))
  "Retuns the minimum element of m."
  (let* ((store (matrix-store m))
	 (max (aref store 0)))
    (declare (type type-blas-store store)
	     (type double-float max))
    (dotimes (i (length store))
      (when (> (aref store i) max)
	(setf max (aref store i))))
    max))

(defmethod mmin ((m matrix-base-dge))
  "Retuns the minimum element of m."
  (let* ((store (matrix-store m))
	 (min (aref store 0)))
    (declare (type type-blas-store store)
	     (type double-float min))
    (dotimes (i (length store))
      (when (< (aref store i) min)
	(setf min (aref store i))))
    min))

(defmethod mminmax ((m matrix-base-dge))
  "Retuns the minimum element of m."
  (let* ((store (matrix-store m))
	 (max (aref store 0))
	 (min (aref store 0)))
    (declare (type type-blas-store store)
	     (type double-float max min))
    (dotimes (i (length store))
      (when (> (aref store i) max)
	(setf max (aref store i)))
      (when (< (aref store i) min)
	(setf min (aref store i))))      
    (list min max)))

(defmethod mabsmax ((m matrix-base-dge))
  "Retuns the minimum element of m."
  (let* ((store (matrix-store m))
	 (max (aref store 0)))
    (declare (type type-blas-store store)
	     (type double-float max))
    (dotimes (i (length store))
      (when (> (abs (aref store i)) (abs max))
	(setf max (aref store i))))
    max))

(defmethod mabsmin ((m matrix-base-dge))
  "Retuns the minimum element of m."
  (let* ((store (matrix-store m))
	 (min (aref store 0)))
    (declare (type type-blas-store store)
	     (type double-float min))
    (dotimes (i (length store))
      (when (< (abs (aref store i)) (abs min))
	(setf min (aref store i))))
    min))

(defmethod .some (pred (a matrix-base-dge) &rest args)
  (let ((stores (mapcar #'matrix-store (cons a args))))
    (apply #'some pred stores)))

(defmethod .every (pred (a matrix-base-dge) &rest args)
  (let ((stores (mapcar #'matrix-store (cons a args))))
    (apply #'every pred stores)))


;;; Matrix and real

(define-constant +generic-function-dfa-df-map+ 
    '((.add . +_dfa-df)
      (.sub . -_dfa-df)
      (.mul . *_dfa-df)
      (.div . /_dfa-df)
      (.expt . ^_dfa-df) 
      (.max . max_dfa-df) 
      (.min . min_dfa-df)))

(defmacro defmethod-dfa-df (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(c (gensym "c")))
  `(defmethod ,name ((,a matrix-base-dge) (,b real))
     (let ((,c (mcreate ,a)))
       (,underlying-function (matrix-store ,a) (coerce ,b 'double-float) (matrix-store ,c))
       ,c))))

(defmacro expand-generic-function-dfa-df-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-dfa-df ,(car x) ,(cdr x)))
	      +generic-function-dfa-df-map+)))

(expand-generic-function-dfa-df-map)

;;; Real and matrix 

(define-constant +generic-function-df-dfa-map+ 
    '((.add . +_df-dfa)
      (.sub . -_df-dfa)
      (.mul . *_df-dfa)
      (.div . /_df-dfa)
      (.expt . ^_df-dfa) 
      (.max . max_df-dfa) 
      (.min . min_df-dfa)))

(defmacro defmethod-df-dfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(c (gensym "c")))
  `(defmethod ,name ((,a real) (,b matrix-base-dge))
     (let ((,c (mcreate ,b)))
       (,underlying-function (coerce ,a 'double-float) (matrix-store ,b) (matrix-store ,c))
       ,c))))

(defmacro expand-generic-function-df-dfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-df-dfa ,(car x) ,(cdr x)))
	      +generic-function-df-dfa-map+)))

(expand-generic-function-df-dfa-map)

;;;; Matrix and matrix

(define-constant +generic-function-dfa-dfa-map+ 
    '((.add . +_dfa-dfa)
      (.sub . -_dfa-dfa)
      (.mul . *_dfa-dfa)
      (.div . /_dfa-dfa)
      (.expt . ^_dfa-dfa) 
      (.max . max_dfa-dfa) 
      (.min . min_dfa-dfa)))

(defmacro defmethod-dfa-dfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(c (gensym "c")))
  `(defmethod ,name ((,a matrix-base-dge) (,b matrix-base-dge))
     (let ((,c (mcreate ,a)))
       (,underlying-function (matrix-store ,a) (matrix-store ,b) (matrix-store ,c))
       ,c))))

(defmacro expand-generic-function-dfa-dfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-dfa-dfa ,(car x) ,(cdr x)))
	      +generic-function-dfa-dfa-map+)))

(expand-generic-function-dfa-dfa-map)




(define-constant +generic-function-dfa-to-dfa-map+ ;really bad name    
    ;; Ordinary functions that map real to real
    ;; Some functions like sqrt are not part of the list 
    ;; since they possibly spit out complex numbers
    '((.sin . sin_dfa-to-dfa) 
      (.cos . cos_dfa-to-dfa) 
      (.tan . tan_dfa-to-dfa) 
      (.atan . atan_dfa-to-dfa)
      (.sinh . sinh_dfa-to-dfa) 
      (.cosh . cosh_dfa-to-dfa) 
      (.tanh . tanh_dfa-to-dfa) 
      (.asinh . asinh_dfa-to-dfa) 
      (.acosh . acosh_dfa-to-dfa) 
      (.exp . exp_dfa-to-dfa) 
      #+nil (.ln . log_dfa) 
      #+nil (.sqrt . sqrt_dfat) 
      #+nil (.conj . conjugate_dfa)
      #+nil (.re . realpart_dfa) 
      #+nil (.im . imagpart_dfa) 
      (.abs . abs_dfa-to-dfa)
      (.sgn . signum_dfa-to-dfa)))

(defmacro defmethod-dfa-to-dfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b")))
  `(defmethod ,name ((,a matrix-base-dge))
     (let ((,b (mcreate ,a)))
       (,underlying-function (matrix-store ,a) (matrix-store ,b) )
       ,b))))

(defmacro expand-generic-function-dfa-to-dfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-dfa-to-dfa ,(car x) ,(cdr x)))
	      +generic-function-dfa-to-dfa-map+)))

(expand-generic-function-dfa-to-dfa-map)

;;; Some exceptions, where output can be either real or complex.

(defmethod .re ((a matrix-base-dge))
  (copy a))

(defmethod .im ((a matrix-base-dge))
  (mcreate a))

(defmethod .conj ((a matrix-base-dge))
  (copy a))

(defmethod .sqrt ((a matrix-base-dge))
  (if (>= (mmin a) 0d0)
      (let ((out (mcreate a)))
	(sqrt_dfa-to-dfa (matrix-store a) (matrix-store out))
	out)
      (let ((out (make-matrix-instance (cons :z (cdr (type-spec a))) 
				       (dim a) 
				       0d0)))
	(sqrt_dfa-to-cdfa (matrix-store a) (matrix-store out))
	out)))
	
(defmethod .ln ((a matrix-base-dge))
  (if (> (mmin a) 0d0)
      (let ((out (mcreate a)))
	(log_dfa-to-dfa (matrix-store a) (matrix-store out))
	out)
      (let ((out (make-matrix-instance (cons :z (cdr (type-spec a))) 
				       (dim a) 
				       0d0)))
	(log_dfa-to-cdfa (matrix-store a) (matrix-store out))
	out)))
	    
(defmethod .asin ((a matrix-base-dge))
  (destructuring-bind (min max) 
      (mminmax a)    
    (if (and (>= min -1d0) 
	     (<= max 1d0))
	(let ((out (mcreate a)))
	  (asin_dfa-to-dfa (matrix-store a) (matrix-store out))
	  out)
	(let ((out (make-matrix-instance (cons :z (cdr (type-spec a))) 
					 (dim a) 
					 0d0)))
	  (asin_dfa-to-cdfa (matrix-store a) (matrix-store out))
	  out))))	

(defmethod .acos ((a matrix-base-dge))
  (destructuring-bind (min max) 
      (mminmax a)    
    (if (and (>= min -1d0) 
	     (<= max 1d0))
	(let ((out (mcreate a)))
	  (acos_dfa-to-dfa (matrix-store a) (matrix-store out))
	  out)
	(let ((out (make-matrix-instance (cons :z (cdr (type-spec a))) 
					 (dim a) 
					 0d0)))
	  (acos_dfa-to-cdfa (matrix-store a) (matrix-store out))
	  out))))

(defmethod .atanh ((a matrix-base-dge))
  (destructuring-bind (min max) 
      (mminmax a)    
    (if (and (> min -1d0) 
	     (< max 1d0))
	(let ((out (mcreate a)))
	  (atanh_dfa-to-dfa (matrix-store a) (matrix-store out))
	  out)
	(let ((out (make-matrix-instance (cons :z (cdr (type-spec a))) 
					 (dim a) 
					 0d0)))
	  (atanh_dfa-to-cdfa (matrix-store a) (matrix-store out))
	  out))))

#|
;;; Now these sad cases where output might be complex for real input

(define-constant +generic-function-dfa-to-cdfa-map+ ;really bad name    
    '((.sqrt . sqrt_dfa)
      (.ln . log_dfa)
      (.asin . asin_dfa)
      (.acos . acos_dfa)
      (.atanh . atanh_dfa)
      ;; Some more?
      )) 

(defmacro defmethod-dfa-to-cdfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(spec (gensym "spec")))
  `(defmethod ,name ((,a matrix-base-dge))
     (let* ((,spec (cons :z (cdr (type-spec ,a))))
	    (,b (make-matrix-instance ,spec (dim ,a) 0)))
       (,underlying-function (matrix-store ,a) (matrix-store ,b) )
       ,b))))

(defmacro expand-generic-function-dfa-to-cdfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-dfa-to-cdfa ,(car x) ,(cdr x)))
	      +generic-function-dfa-to-cdfa-map+)))

(expand-generic-function-dfa-to-cdfa-map)
|#