;;; Lisplab, level2-matrix-zge.lisp
;;; Optimizations for complex matrices.

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License along
;;; with this program; if not, write to the Free Software Foundation, Inc.,
;;; 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

(in-package :lisplab)

(defmethod mfill ((a matrix-base-zge) value)
  (let ((rx (coerce (realpart value) 'double-float))
	(cx (coerce (imagpart value) 'double-float))
	(store (matrix-store a)))
    (declare (type type-blas-store store))
    (loop for i from 0 below (length store) by 2 do 
      (setf (aref store i) rx
	    (aref store (1+ i)) cx))))

(defmethod copy  ((matrix matrix-base-zge))
  (make-instance (class-name (class-of matrix))
		 :store (copy-seq (the type-blas-store (matrix-store matrix)))
		 :dim (dim matrix)))

(defmethod copy-contents ((a matrix-base-zge) (b matrix-base-zge) 
			  &optional (converter nil))
  (let ((store-a (matrix-store a))
	(store-b (matrix-store b)))
    (if converter 
	(map-into store-b converter store-a)
	(copy-matrix-stores store-a store-b)))    
    b)

(defmethod copy-contents ((from matrix-base-dge) (to matrix-base-zge) 
			  &optional (converter nil))
  (if converter 
      (call-next-method) ;; Could have some testes here to improve performance
      (let* ((store-a (matrix-store from))
	     (store-b (matrix-store to))
	     (len (length store-a)))
	(declare (type type-blas-store store-a store-b)
		 (type type-blas-idx len))
	(dotimes (i len)
	  (setf (aref store-b (truly-the type-blas-idx (* 2 i))) (aref store-a i)))
	to))) 

(defmethod msum ((m matrix-base-zge))
  (let ((sum-r 0d0)
        (sum-i 0d0)
        (m0 (matrix-store m)))
    (declare (type double-float sum-r sum-i)
             (type type-blas-store m0))
    (loop for i from 0 below (length m0) by 2 do
          (incf sum-r (aref m0 i))
          (incf sum-i (aref m0 (1+ i))))
    (complex sum-r sum-i)))


 ;;; Matrix and complex

(define-constant +generic-function-cdfa-cdf-map+ 
    '((.add . +_cdfa-cdf)
      (.sub . -_cdfa-cdf)
      (.mul . *_cdfa-cdf)
      (.div . /_cdfa-cdf)
      (.expt . ^_cdfa-cdf)))

(defmacro defmethod-cdfa-cdf (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(c (gensym "c")))
  `(defmethod ,name ((,a matrix-base-zge) (,b number))
     (let ((,c (mcreate ,a)))
       (,underlying-function (matrix-store ,a) 
			     (coerce ,b '(complex double-float)) 
			     (matrix-store ,c))
       ,c))))

(defmacro expand-generic-function-cdfa-cdf-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-cdfa-cdf ,(car x) ,(cdr x)))
	      +generic-function-cdfa-cdf-map+)))

(expand-generic-function-cdfa-cdf-map)

;;; Complex and matrix 

(define-constant +generic-function-cdf-cdfa-map+ 
    '((.add . +_cdf-cdfa)
      (.sub . -_cdf-cdfa)
      (.mul . *_cdf-cdfa)
      (.div . /_cdf-cdfa)
      (.expt . ^_cdf-cdfa)))

(defmacro defmethod-cdf-cdfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(c (gensym "c")))
  `(defmethod ,name ((,a number) (,b matrix-base-zge))
     (let ((,c (mcreate ,b)))
       (,underlying-function (coerce ,a '(complex double-float)) 
			     (matrix-store ,b) 
			     (matrix-store ,c))
       ,c))))

(defmacro expand-generic-function-cdf-cdfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-cdf-cdfa ,(car x) ,(cdr x)))
	      +generic-function-cdf-cdfa-map+)))

(expand-generic-function-cdf-cdfa-map)

;;;; Matrix and matrix

(define-constant +generic-function-cdfa-cdfa-map+ 
    '((.add . +_cdfa-cdfa)
      (.sub . -_cdfa-cdfa)
      (.mul . *_cdfa-cdfa)
      (.div . /_cdfa-cdfa)
      (.expt . ^_cdfa-cdfa) ))

(defmacro defmethod-cdfa-cdfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b"))
	(c (gensym "c")))
  `(defmethod ,name ((,a matrix-base-zge) (,b matrix-base-zge))
     (let ((,c (mcreate ,a)))
       (,underlying-function (matrix-store ,a) (matrix-store ,b) (matrix-store ,c))
       ,c))))

(defmacro expand-generic-function-cdfa-cdfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-cdfa-cdfa ,(car x) ,(cdr x)))
	      +generic-function-cdfa-cdfa-map+)))

(expand-generic-function-cdfa-cdfa-map)

;;;; The cross terms, where one input is real, the other complex

(defmacro def-cross-complex-real-method (name type1 type2)
  ;; Assumes output type same as input
  ;; Note will crash if input is not converted properly!!!!!!
  (let ((a (gensym))
	(b (gensym))
	(spec (gensym)))
    `(defmethod ,name ((,a ,type1) (,b ,type2))
       ;; This is not the fastest possible, but rather fast
       (let ((,spec (operator-output-type-spec ',name 
					       (type-spec ,a)
					       (type-spec ,b))))
	 ;; Assumes that input is something with a well-defined spec
	 (,name (if (matrix-p ,a) (convert ,a ,spec) ,a)  
		(if (matrix-p ,b) (convert ,b ,spec) ,b))))))
			 
(defmacro def-all-cross-complex-real-methods (name)
  `(progn
     (def-cross-complex-real-method ,name complex matrix-base-dge)
     (def-cross-complex-real-method ,name matrix-base-dge complex)
   ;;  (def-cross-complex-real-method ,name matrix-base-zge matrix-base-dge)
   ;;  (def-cross-complex-real-method ,name matrix-base-dge matrix-base-zge)
     'done))

(def-all-cross-complex-real-methods .add)
(def-all-cross-complex-real-methods .sub)
(def-all-cross-complex-real-methods .mul)
(def-all-cross-complex-real-methods .div)
(def-all-cross-complex-real-methods .expt)

;;; Add

(defmethod .add ((a matrix-base-zge) (b real))
  (let ((c (mcreate a)))
    (+_cdfa-df (matrix-store a) (coerce b 'double-float) (matrix-store c))
    c))

(defmethod .add ((b real) (a matrix-base-zge) )
  (.add a b))

(defmethod .add ((a matrix-base-zge) (b matrix-base-dge))
  (let ((c (mcreate a)))
    (+_cdfa-dfa (matrix-store a) (matrix-store b) (matrix-store c))
    c))

(defmethod .add ((b matrix-base-dge) (a matrix-base-zge))
  (.add a b))

;;; Sub

(defmethod .sub ((a matrix-base-zge) (b real))
  (let ((c (mcreate a)))
    (-_cdfa-df (matrix-store a) (coerce b 'double-float) (matrix-store c))
    c))

(defmethod .sub ((a real) (b matrix-base-zge))
  (let ((c (mcreate b)))
    (-_df-cdfa (coerce a 'double-float) (matrix-store b) (matrix-store c))
    c))

(defmethod .sub ((a matrix-base-zge) (b matrix-base-dge))
  (let ((c (mcreate a)))
    (-_cdfa-dfa (matrix-store a) (matrix-store b) (matrix-store c))
    c))

(defmethod .sub ((a matrix-base-dge) (b matrix-base-zge))
  (let ((c (mcreate b)))
    (-_dfa-cdfa (matrix-store a) (matrix-store b) (matrix-store c))
    c))

;;; Mul

(defmethod .mul ((a matrix-base-zge) (b real))
  (let ((c (mcreate a)))
    (*_cdfa-df (matrix-store a) (coerce b 'double-float) (matrix-store c))
    c))

(defmethod .mul ((b real) (a matrix-base-zge) )
  (.mul a b))

(defmethod .mul ((a matrix-base-zge) (b matrix-base-dge))
  (let ((c (mcreate a)))
    (*_cdfa-dfa (matrix-store a) (matrix-store b) (matrix-store c))
    c))

(defmethod .mul ((b matrix-base-dge) (a matrix-base-zge))
  (.mul a b))

;;; Div

(defmethod .div ((a matrix-base-zge) (b real))
  (let ((c (mcreate a)))
    (/_cdfa-df (matrix-store a) (coerce b 'double-float) (matrix-store c))
    c))

(defmethod .div ((a matrix-base-zge) (b matrix-base-dge))
  (let ((c (mcreate a)))
    (/_cdfa-dfa (matrix-store a) (matrix-store b) (matrix-store c))
    c))

(def-cross-complex-real-method .div matrix-base-dge matrix-base-zge)

;;; Expt

(def-cross-complex-real-method .expt matrix-base-zge matrix-base-dge)
(def-cross-complex-real-method .expt matrix-base-dge matrix-base-zge)


;;;; Ordinary functions

(define-constant +generic-function-cdfa-to-cdfa-map+ ;really bad name    
    '((.sin . sin_cdfa-to-cdfa) 
      (.cos . cos_cdfa-to-cdfa) 
      (.tan . tan_cdfa-to-cdfa) 
      (.asin . asin_cdfa-to-cdfa)   
      (.acos . acos_cdfa-to-cdfa)  
      (.atan . atan_cdfa-to-cdfa) 
      (.sinh . sinh_cdfa-to-cdfa) 
      (.cosh . cosh_cdfa-to-cdfa) 
      (.tanh . tanh_cdfa-to-cdfa) 
      (.asinh . asinh_cdfa-to-cdfa) 
      (.acosh . acosh_cdfa-to-cdfa) 
      (.atanh . atanh_cdfa-to-cdfa)
      (.exp . exp_cdfa-to-cdfa) 
      (.ln . log_cdfa-to-cdfa) 
      (.sqrt . sqrt_cdfa-to-cdfa) 
      (.conj . conjugate_cdfa-to-cdfa)))

(defmacro defmethod-cdfa-to-cdfa (name underlying-function)  
  (let ((a (gensym "a"))
	(b (gensym "b")))
  `(defmethod ,name ((,a matrix-base-zge))
     (let ((,b (mcreate ,a)))
       (,underlying-function (matrix-store ,a) (matrix-store ,b) )
       ,b))))

(defmacro expand-generic-function-cdfa-to-cdfa-map ()
  (cons 'progn
      (mapcar (lambda (x)
		`(defmethod-cdfa-to-cdfa ,(car x) ,(cdr x)))
	      +generic-function-cdfa-to-cdfa-map+)))

(expand-generic-function-cdfa-to-cdfa-map)

;;;; Exceptions, in that output is real for complex input

(defmethod .im ((a matrix-base-zge))
  (let* ((spec-out (cons :d (cdr (type-spec a))))
	 (out (make-matrix-instance spec-out (dim a) 0)))
    (imagpart_cdfa (matrix-store a) (matrix-store out))
    out))

(defmethod .re ((a matrix-base-zge))
  (let* ((spec-out (cons :d (cdr (type-spec a))))
	 (out (make-matrix-instance spec-out (dim a) 0)))
    (realpart_cdfa (matrix-store a) (matrix-store out))
    out))

(defmethod .abs ((a matrix-base-zge))
  (let* ((spec-out (cons :d (cdr (type-spec a))))
	 (out (make-matrix-instance spec-out (dim a) 0)))
    (abs_cdfa (matrix-store a) (matrix-store out))
    out))
	



